import os
import subprocess
import sys

import click
from termcolor import colored
from src.tmux import (
    get_current_window,
    clean_tmux_window,
    is_inside_tmux,
    setup_tmux_client,
)


@click.command(help="Proxy for yarn kbn bootstrap")
@click.option("--clean", "-c", is_flag=True, help="Runs bootstrap with clean")
@click.option(
    "--tmux",
    is_flag=True,
    help=(
        "If you run kibbe inside a tmux session it will divide the current window, run"
        " elasticsearch (kibbe es) and kibana (kibbe kibana)"
    ),
)
@click.option(
    "--flush",
    is_flag=True,
    default=False,
    help="If passed will pass the flush flag to ES when initializing",
)
def setmeup(clean, tmux, flush):

    if tmux and not is_inside_tmux():
        setup_tmux_client()
        exit()

    commands = []
    if clean:
        commands.append(["yarn", "kbn", "bootstrap", "clean"])

    commands.append(["yarn", "kbn", "bootstrap"])

    for command in commands:
        subprocess.run(command)

    if tmux:
        setup_tmux(flush=flush)


def setup_tmux(flush=False):

    current_window = get_current_window()
    clean_tmux_window(current_window)
    current_pane_id = os.getenv("TMUX_PANE")
    current_pane = current_window.get_by_id(current_pane_id)

    es_pane = current_window.split_window(vertical=False)
    start_es(es_pane, flush=flush)

    click.echo(colored("Starting kibana and closing this pane", "yellow"))
    kibana_pane = current_window.split_window(vertical=False)
    start_kibana(kibana_pane)

    # close this pane. This will also kill the python process
    current_pane.cmd("kill-pane")
    exit()


def start_kibana(kibana_pane):
    current_dir = os.getcwd()
    kibbe_command = sys.argv[0]

    kibana_pane.send_keys("cd %s" % current_dir)
    kibana_pane.send_keys("nvm use")
    kibana_pane.send_keys("%s kibana --wait" % kibbe_command)


def start_es(es_pane, flush=False):
    current_dir = os.getcwd()
    kibbe_command = sys.argv[0]

    params = ""
    if flush:
        params = "--flush"

    es_pane.send_keys("cd %s" % current_dir)
    es_pane.send_keys("nvm use")
    es_pane.send_keys("%s es %s" % (kibbe_command, params))



import click
import atexit
import psutil

from src.config import get_config, make_config_files
from src.commands.check import check
from src.commands.config import config
from src.commands.ctx import ctx
from src.commands.es import es
from src.commands.fleet import fleet
from src.commands.kibana import kibana
from src.commands.setlicense import setlicense
from src.commands.setmeup import setmeup
from src.commands.jest import jest
from src.commands.checkout import checkout
from src.util import force_kibana_root


@click.group(
    help="""
        Kibbe is a tool that help with common tasks when developing kibana plugins.

        Some subcommands allow you to define a configuration to persist arguments for some
        specific tasks such as running kibana or elasticsearch. This configuration file will
        persis those arguments among all kibana clones and branches.

        You can create a configuration file in your home ~/.kibbe and follow the configuration
        example in the kibbe repository
"""
)
def cli():
    force_kibana_root()
    config = get_config()
    make_config_files(config)


cli.add_command(setlicense)
cli.add_command(check)
cli.add_command(es)
cli.add_command(kibana)
cli.add_command(setmeup)
cli.add_command(fleet)
cli.add_command(ctx)
cli.add_command(config)
cli.add_command(jest)
cli.add_command(checkout)


def exit_():
    """
    Makes sure that when exiting kibbe any remaining subprocess is killed.
    This is useful because if kibbe starts a nodejs process it might spawn
    more sub-proceses but they will not be terminated if the parent is asked to
    do so.
    """
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    for child in children:
        try:
            child.terminate()
        except Exception:
            pass


atexit.register(exit_)



import click
import os
import subprocess
import re

from termcolor import colored

from src.config import get_kibbe_config


@click.command()
@click.argument("testfile", type=click.Path(exists=True), required=False)
@click.option(
    "--watch", "-w", default=False, is_flag=True, help="Run jest in watch mode"
)
def jest(testfile, watch):
    """
    Helper for kibana's jest runner script.

    If TESTFILE is passed. It will run jest for it.

    If TESTFILE is not passed it will prompt you to select a test file
    from `git status`
    """
    if testfile:
        run_for_file(testfile, watch)
        return

    file_re = re.compile(r"\.test\.(t|j)sx?$")

    status = subprocess.getoutput("git status --porcelain")
    modifiedlist = status.split("\n")

    selectable = []
    for entry in modifiedlist:
        try:
            file = entry.split(" ")[-1]
            if file_re.search(file):
                selectable.append(file)
        except Exception:
            pass

    if len(selectable) == 0:
        click.echo("No test files modified")
        return

    counter = 1
    choices = []
    click.echo(colored("Select a file to test", "blue"))
    for file in selectable:
        click.echo("%d. %s" % (counter, file))
        choices.append(str(counter))
        counter = counter + 1

    file_nr = click.prompt(
        text="File list number",
        default="1",
        show_choices=False,
        type=click.Choice(choices),
    )

    run_for_file(selectable[int(file_nr) - 1], watch)


def run_for_file(testfile: str, watch: bool):
    command_args = ""
    max_workers = get_kibbe_config("jest-max-workers")

    if max_workers:
        command_args = command_args + " --maxWorkers=%s" % max_workers

    if watch:
        command_args = command_args + " --watch"

    command = "node scripts/jest.js %s %s" % (command_args, testfile)
    click.echo(colored(command, "yellow"))
    os.system("node scripts/jest.js %s %s" % (command_args, testfile))



import os
from pathlib import Path
from src.git import find_existing_worktree, get_worktree_list_flat
from src.tmux import get_current_panel, is_inside_tmux
import subprocess

import click
from termcolor import colored

from src.util import get_valid_filename


@click.command()
@click.argument(
    "name", type=click.STRING, shell_complete=get_worktree_list_flat, required=False
)
@click.option(
    "--branch",
    help="Branch name to use for the new worktree",
    default="",
)
@click.option(
    "--parent-path",
    help="Custom parent path to use to set the new worktree. Defaults to ../",
    default="",
)
@click.option(
    "--source",
    help="Branch to create the worktree from. Defaults to current branch. e.g. master",
    default="",
)
@click.option(
    "-i",
    "--interactive",
    help="Ask questions about the new context",
    is_flag=True,
    default=True,
)
@click.option(
    "-B",
    "--overwrite-branch",
    help="If the branch already exists, reset it to source",
    is_flag=True,
    default=False,
)
@click.option(
    "--cd",
    "--change-dir",
    help="Change to new context directory (Tmux only)",
    is_flag=True,
    default=True,
)
@click.option(
    "--delete",
    help="Removes the worktree if exists",
    is_flag=True,
    default=False,
    confirmation_prompt=True,
)
@click.option(
    "-l",
    "--list-worktrees",
    help="List existing worktree. Alias for `git worktree list`",
    is_flag=True,
    default=False,
)
def ctx(
    name,
    parent_path,
    source,
    branch,
    interactive,
    overwrite_branch,
    cd,
    list_worktrees,
    delete,
):
    """
    ctx is a wrapper kibbe subcommand for git worktree with some quality of life improvements.

    NAME accepts a name of the "context" you want to switch. It is a shorthand to not have to remember
    paths as is required with git worktree.

    It allows you to quickly switch and create git worktrees without having to type or memorize
    all the git worktree parameterse.

    ctx works better when you use it with tmux. In mac, if you use iterm2,
    you can start tmux with `tmux -CC`. Install it first with `brew install tmux`

    ctx is not intended to be a replacement for git worktree, if you can't perform the operation
    you want with ctx please see the git worktree manual entry https://git-scm.com/docs/git-worktree
    """

    if list_worktrees:
        subprocess.run(["git", "worktree", "list"])
        exit(0)

    if not name:
        raise click.ClickException(
            colored(
                "You must pass the NAME of the worktree you want to change to", "red"
            )
        )

    if name.startswith("../"):
        name = name[3:]

    path_name = get_valid_filename(name)
    existing_worktree = find_existing_worktree(path_name)

    if delete and not existing_worktree:
        raise click.ClickException(
            "Can not remove worktree. Worktree doesn't exist: "
            + colored(path_name, "red")
        )
    elif delete:
        if click.confirm(
            "Are you sure you want to delete the wortree"
            + colored(existing_worktree["worktree"], "yellow")
        ):
            click.echo("Deleting worktree...")
            subprocess.run(["git", "worktree", "remove", existing_worktree["worktree"]])
        exit(0)

    if existing_worktree:
        return handle_existing_worktree(existing_worktree)

    if not branch:
        if interactive:
            branch = click.prompt("Git branch name fore the new worktree", default=name)
        else:
            branch = name

    if not parent_path:
        possible_path = os.path.join(Path(os.getcwd()).parent.absolute())
        if interactive:
            parent_path = click.prompt(
                "Parent path target for the git worktree",
                default=possible_path,
                type=click.Path(dir_okay=True, file_okay=False),
            )
        else:
            parent_path = possible_path

    full_path = os.path.join(parent_path, path_name)

    if not source:
        possible_source = subprocess.getoutput("git rev-parse --abbrev-ref HEAD")
        if interactive:
            source = click.prompt(
                "Source branch for the git worktree. e.g. master",
                default=possible_source,
            )
        else:
            source = possible_source

    click.echo("Will create a new git worktree called: " + colored(path_name, "yellow"))
    click.echo("In this location: " + colored(parent_path, "blue"))
    click.echo("With a new branch name: " + colored(branch, "blue"))
    click.echo("From this branch: " + colored(source, "blue"))
    click.echo("---git output--")

    b_option = "-b" if not overwrite_branch else "-B"

    command = ["git", "worktree", "add", full_path, source, b_option, branch]
    process = subprocess.run(command)
    click.echo("--- end git output---")

    if process.returncode != 0:
        raise click.ClickException(
            colored(
                "Something went wrong with git. See git output and verify your"
                " parameters",
                "red",
            )
        )

    click.echo(
        colored("Success!", "green")
        + " a new git worktree was created in "
        + colored(full_path, "blue")
    )

    click.echo("To change to your new worktree run:")
    click.echo(colored("cd %s" % full_path, "yellow"))

    # this must always be the last command
    if cd and is_inside_tmux():
        click.echo("Tmux session detected. Changing to worktree")
        current_pane = get_current_panel()
        current_pane.send_keys("cd %s && nvm use" % full_path)
        exit(0)
    elif not is_inside_tmux():
        click.echo(
            "Changing to a worktree is only supported if you are running inside tmux"
        )


def handle_existing_worktree(existing_worktree):
    existing_path_name = existing_worktree["worktree"]
    click.echo(
        "Existing worktree with the same name found at "
        + colored(existing_path_name, "yellow")
    )
    click.echo("Worktree branch: " + colored(existing_worktree["branch"], "blue"))
    click.echo("Head commit: " + colored(existing_worktree["HEAD"], "blue"))
    click.echo()

    if Path(existing_path_name) == Path(os.getcwd()):
        click.echo(colored("You are already on this worktree", "yellow"))
        exit(0)

    if not is_inside_tmux():
        click.echo("You can switch to it by running: ")
        click.echo(colored("cd %s" % existing_path_name, "blue"))
        click.echo()
        click.echo("Run this command inside tmux to automatically cd to it")
    else:
        click.echo("Tmux session detected. Changing to worktree")
        current_pane = get_current_panel()
        current_pane.send_keys("cd %s && nvm use" % existing_path_name)
        exit(0)



import click
from termcolor import colored

from src.config import get_config, print_config


@click.command("config")
def config():
    """
    Prints the current kibbe configuration and where is it taking it from
    """
    click.echo(colored("Configuration to use", "yellow"))
    config = get_config()
    print_config(config)



import os
from pathlib import Path
import subprocess


def get_worktree_list():
    try:
        raw_list = subprocess.getoutput("git worktree list --porcelain")
        worktrees = []
        raw_list = raw_list.split("\n")

        current = {}
        for item in raw_list:
            if not item:
                worktrees.append(current)
                current = {}
                continue

            [name, value] = item.split(" ")
            current[name] = value

        return worktrees
    except ValueError:
        return []


def get_worktree_list_flat(ctx, param, incomplete=""):
    final = []
    try:
        worktrees = get_worktree_list()
        for tree in worktrees:
            name = os.path.basename(tree["worktree"])
            if not incomplete or (incomplete and name.startswith(incomplete)):
                final.append(name)
    except ValueError:
        pass

    return final


def find_existing_worktree(path_name):
    worktrees = get_worktree_list()
    existing_worktree = {}
    for tree in worktrees:
        path = Path(tree["worktree"])
        if path.name == path_name:
            existing_worktree = tree
            break

    return existing_worktree



import json
import os
from pathlib import PurePath
import re
import subprocess
import time

import click
import requests


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""

    # from whichcraft import which
    from shutil import which

    return which(name) is not None


def is_es_running():
    try:
        requests.get("http://localhost:9200")
        return True
    except requests.ConnectionError:
        return False


def is_kibana_running():
    try:
        requests.get("http://localhost:5601")
        return True
    except requests.ConnectionError:
        return False


def force_kibana_root():
    if not is_kibana_repo():
        raise click.ClickException(
            "You must run this command in the root of a kibana repo clone"
        )


def is_kibana_repo():
    if not os.path.isfile("package.json"):
        return False

    file = open("package.json")
    try:
        content = json.load(file)
        if content["name"] != "kibana" or not content["homepage"].startswith(
            "https://www.elastic.co/"
        ):
            return False
    except ValueError:
        return False
    finally:
        file.close()

    return True


def get_modified_files():
    files = ""
    try:
        files = subprocess.getoutput("git diff --name-only HEAD")
    except ValueError:
        return []

    files = filter(None, files.split("\n"))
    return list(files)


def find_related_test(file):
    path = PurePath(file)

    # skip if the file is a test
    if path.match("*.test.*"):
        return ""

    test_file = path.with_suffix(".test" + path.suffix)

    if os.path.isfile(test_file):
        return test_file

    return ""


def find_related_plugin_folder(file):
    path = PurePath(file)

    try:
        if not path.relative_to("x-pack/plugins"):
            return ""
    except ValueError:
        return ""

    while not path.match("x-pack/plugins/*"):
        path = PurePath(path.parent)

    return str(path)


def merge_params(config_params, unparsed_args, useEqual=False):
    final_params = []
    params_map = {}
    for conf, value in config_params:
        params_map["--" + conf] = value

    skip = False
    for index, param in enumerate(unparsed_args):
        if skip:
            skip = False
            continue
        nextIsValue = len(unparsed_args) > index + 1 and not str(
            unparsed_args[index + 1]
        ).startswith("--")
        if param in params_map and nextIsValue:
            params_map[param] = unparsed_args[index + 1]
            skip = True
        else:
            params_map[param] = unparsed_args[index + 1] if nextIsValue else ""
            if nextIsValue:
                skip = True

    if useEqual:
        for param in params_map:
            if len(params_map[param]) > 0:
                final_params.append(param + "=" + params_map[param])
            else:
                final_params.append(param)
    else:
        for param in params_map:
            final_params.append(param)
            if len(params_map[param]) > 0:
                final_params.append(params_map[param])

    return final_params


def unparsed_to_map(params):
    params_map = {}
    skip = False
    for index, param in enumerate(params):
        if skip:
            skip = False
            continue
        nextIsValue = len(params) > index + 1 and not str(params[index + 1]).startswith(
            "--"
        )
        if param.startswith("--"):
            if nextIsValue:
                params_map[param] = params[index + 1]
                skip = True
            else:
                params_map[param] = ""
    return params_map


def wait_for_elastic_search():
    total = 60
    current = total
    numbers = list(range(1, total))
    with click.progressbar(numbers) as bar:
        for item in bar:
            current = item
            if is_es_running():
                break
            time.sleep(1)

    if current <= 0:
        return True
    else:
        return False


def get_valid_filename(name):
    s = str(name).strip().replace(" ", "_")
    s = re.sub(r"(?u)[^-\w.]", "-", s)
    return s



import os
from pathlib import PurePath
import subprocess

import click
from termcolor import colored

from src.util import find_related_plugin_folder, find_related_test, get_modified_files


@click.command(
    help="""Runs quick checks for your modified files. Useful to run before committing"""
)
@click.option("--skip-tests", default=False, help="Skip running tests", is_flag=True)
@click.option(
    "--skip-stylescheck", default=False, help="Skip running styles check", is_flag=True
)
@click.option(
    "--types",
    default=False,
    help="Runs Typescript types check (if relevant)",
    is_flag=True,
)
def check(skip_tests, types, skip_stylescheck):

    modified = get_modified_files()

    if len(modified) == 0:
        click.echo(
            colored(
                "Kibbe check only works with uncommited files. Edit some files first",
                "yellow",
            )
        )

    configs = set()
    tests = set()
    styles = set()
    for file in modified:
        path = PurePath(file)
        if path.suffix in [".css", ".scss"]:
            styles.add(file)
        tests.add(find_related_test(file))
        configs.add(find_related_plugin_folder(file))

    tests = list(filter(None, tests))
    configs = list(filter(None, configs))
    styles = list(filter(None, styles))

    fails = []

    # style check
    if not skip_stylescheck and len(styles) > 0:
        click.echo(colored(">> Running styles check", "yellow"))
        process = subprocess.run(["node", "scripts/stylelint.js"] + styles)
        if process.returncode != 0:
            fails.append("Stylelint")

    # types check
    if types and len(configs) > 0:
        click.echo(colored(">> Finding tsconfig.json", "yellow"))
        for config in configs:
            tsconfig = os.path.join(config, "tsconfig.json")
            if os.path.isfile(tsconfig):
                click.echo(
                    colored(">>  - Running type_check for ", "cyan")
                    + colored(tsconfig, "white")
                )
                process = subprocess.run(
                    ["node", "scripts/type_check.js", "--project", tsconfig]
                )
                if process.returncode != 0:
                    fails.append("TS types check")

    # eslint
    if len(modified) > 0:
        click.echo(colored(">> Running eslint", "yellow"))
        process = subprocess.run(["node", "scripts/eslint.js"] + modified)
        if process.returncode != 0:
            fails.append("Eslint checks")

    # tests
    if not skip_tests and len(tests) > 0:
        click.echo(colored(">> Running related tests", "yellow"))
        process = subprocess.run(["node", "scripts/jest.js"] + tests)
        if process.returncode != 0:
            fails.append("Jest tests")

    if len(fails) > 0:
        click.echo("\n ------ \n")
        click.echo(colored(">> Errors found running checks:\n", "red"))
        for fail in fails:
            click.echo("      ❌ - " + str(fail))

    click.echo("kibbe check complete")



import subprocess
import ipaddress
from termcolor import colored
from src.util import is_es_running, is_kibana_running
import click
import re

_RE_COMBINE_WHITESPACE = re.compile(r"\s+")


@click.command()
@click.option(
    "--snapshot",
    default="docker.elastic.co/beats/elastic-agent:8.0.0-SNAPSHOT",
    help=(
        "If you want to specify a different snapshot. Default to"
        " docker.elastic.co/beats/elastic-agent:8.0.0-SNAPSHOT "
    ),
)
@click.option(
    "--docker-ip",
    help=(
        "The docker host IP. Kibbe will try to autodetect it but you can overwrite it"
        " with this option"
    ),
)
@click.option(
    "--run",
    is_flag=True,
    help=(
        "It will attempt to run fleet server locally. You should have kibana and"
        " elastic search running locally already"
    ),
)
def fleet(run, snapshot, docker_ip):
    """
    Utilities to work with the security solutions fleet server locally
    """
    if run:
        if not is_kibana_running():
            click.echo(
                colored(
                    "Kibana is not running. Kibana must be runnnig to run fleet via"
                    " this helper",
                    "red",
                )
            )
            raise click.ClickException("Kibana not running")
        if not is_es_running():
            click.echo(
                colored(
                    "Elasticsearch is not running. Elasticsearch  must be runnnig to"
                    " run fleet via this helper",
                    "red",
                )
            )
            raise click.ClickException("Elasticsearch not running")

        docker_ip = get_docker_ip(docker_ip)
        click.echo(" - Autodetected docker host ip: " + colored(docker_ip, "blue"))

        docker_command = """docker run \
            --restart no \
            --add-host kibana:{host_ip} \
            --add-host elasticsearch:{host_ip} \
            --add-host fleetserver:127.0.0.1 \
            -e KIBANA_HOST=http://kibana:5601 \
            -e KIBANA_USERNAME=elastic \
            -e KIBANA_PASSWORD=changeme \
            -e ELASTICSEARCH_HOST=http://elasticsearch:9200 \
            -e ELASTICSEARCH_USERNAME=elastic  \
            -e ELASTICSEARCH_PASSWORD=changeme \
            -e FLEET_SERVER_HOST=0.0.0.0 \
            -e FLEET_INSECURE=1 \
            -e KIBANA_FLEET_SETUP=1 \
            -e FLEET_SERVER_ENABLE=1 \
            -e FLEET_SERVER_INSECURE_HTTP=1 \
            -p 8220:8220 {snapshot}
        """.replace(
            "\n", ""
        )
        docker_command = docker_command.format(host_ip=docker_ip, snapshot=snapshot)
        docker_command = _RE_COMBINE_WHITESPACE.sub(" ", docker_command).strip()

        click.echo(" - Will run docker with:\n")
        click.echo(colored(docker_command, "yellow"))
        docker_command = docker_command.split(" ")
        subprocess.run(docker_command)


def get_docker_ip(default_ip):
    ip = default_ip if default_ip and len(default_ip > 0) else "172.17.0.1"
    possible_ip = subprocess.getoutput(
        "docker network inspect bridge -f '{{range .IPAM.Config}}{{.Gateway}}{{end}}'"
    )
    try:
        ipaddress.ip_address(possible_ip)
        ip = possible_ip
    except ValueError:
        pass

    return ip



import subprocess
import sys

import click
import re
import enlighten
from termcolor import colored

from src.config import get_config, get_kibbe_config, persist_config
from src.util import is_es_running, merge_params, unparsed_to_map
from src.util import wait_for_elastic_search


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--wait",
    "-w",
    default=False,
    is_flag=True,
    help=(
        "If passed. It will wait for an elastic search instance in the default port"
        " (9200) to be ready before starting kibana"
    ),
)
@click.option(
    "--alt",
    default=False,
    is_flag=True,
    help="Shows an alterantive kibana loading log. Based on text parsing and regex.",
)
@click.option(
    "--prod",
    "-p",
    default=False,
    is_flag=True,
    help="Runs Kibana in production mode (omits the '--dev' argument).",
)
@click.option(
    "--save-config",
    default=False,
    is_flag=True,
    help=(
        "If passed it will write your kibbe configuration with all the current passed"
        " parameters. This will not modify your kibana repo clone."
    ),
)
@click.argument("unparsed_args", nargs=-1, type=click.UNPROCESSED)
def kibana(save_config, unparsed_args, wait, alt, prod):
    """
    Runs Kibana from the current clone.

    You can pass the same parameters as you'd pass to `node scripts/kibana`

    You can persist some parameters by using a configuration file `~/.kibbe`.
    with the [kibana.params] section.

    See more about the configuration file here:
    https://github.com/academo/kibbe#configuration-file
    """

    if wait and not is_es_running():
        click.echo(
            colored("Waiting for elasticsearch in port 9200. Timeout in 60s", "blue")
        )
        wait_for_elastic_search()
    config_params = []
    config = get_config()

    if "kibana.params" in config:
        config_params = config.items("kibana.params", raw=True)

    params = merge_params(config_params, unparsed_args, useEqual=True)

    if save_config:
        persist_config({"kibana.params": unparsed_to_map(params)})
        exit()

    command = ["node", "scripts/kibana"]
    if not prod:
        command.append("--dev")
    command.extend(params)
    click.echo("Will run kibana search as: " + colored(" ".join(command), "yellow"))

    if alt or get_kibbe_config("kibana-alt-mode"):
        run_kibana_alt(command)
    else:
        subprocess.run(command)


def run_kibana_alt(command):
    # command = ["node test.js"]

    manager = enlighten.get_manager()

    pbar = manager.counter(
        total=100, desc="Optimizer", unit="bundles", color="blue_on_green"
    )
    status = manager.status_bar(
        fields={"kibana_status": "Initializing"},
        status_format="Kibana is {kibana_status}",
        color="white_on_black",
    )

    process = subprocess.Popen(
        "FORCE_COLOR=1 " + " ".join(command),
        shell=True,
        stdout=subprocess.PIPE,
    )

    pbar.count = int(0)
    pbar.refresh()

    if process.stdout:
        while True:
            # exit kibbe if the node process died
            if process.poll() is not None or process.returncode:
                sys.exit()
            output = process.stdout.readline()
            if output:
                line = output.decode("utf-8")
                sys.stdout.write(line)
                parse_line(line, pbar, status)


optimizerProgressRe = re.compile(r"^.*?@kbn\/optimizer.*?\[(\d+)\/(\d+)\]\s.*$")
optimizerSuccessRe = re.compile(r"^.*?success.*?kbn\/optimizer.*")

kibanaServerRunning = re.compile(r".*http\.server\.Kibana.*?http server running")
kibanaServerStatus = re.compile(r".*?status.*?\sKibana\sis\snow\s(.+)(?:\s|$)")


def parse_line(line: str, pbar, status: enlighten.StatusBar):
    progressMatch = optimizerProgressRe.match(line)
    if progressMatch:
        current = int(progressMatch.group(1))
        total = int(progressMatch.group(2))
        pbar.total = total
        pbar.count = current
        pbar.refresh()
        return
    successMatch = optimizerSuccessRe.match(line)
    if successMatch:
        pbar.clear()
        return

    if kibanaServerRunning.match(line):
        status.fields["kibana_status"] = "⌛ Server loading"
        status.refresh()
        return

    kibanaStatusMatch = kibanaServerStatus.match(line)
    if kibanaStatusMatch:
        message = str(kibanaStatusMatch.group(1))
        message = get_kibana_icon(message) + message
        status.fields["kibana_status"] = message
        status.refresh()
        return


def get_kibana_icon(message):
    if "available" in message:
        return "✅ "

    if "degraded" in message:
        return "⌛ "

    return ""



import click
import subprocess
from src.util import is_tool

valid_licenses = ['trial', 'basic', 'gold', 'platinum', 'enterprise']


@click.command(help="""This command is a proxy for @pzl setlicense tool
        https://github.com/elastic/pzl-es-tools""")
@click.argument('license_type')
@click.option('-e', '--elastic', default="http://localhost:9200",
              help="Elastic search host. Default to localhost:9200")
@click.option('-p', '--password', default="changeme",
              help="Password to connect. Default to changeme")
@click.option('-u', '--user', default="elastic",
              help="Username to connect. Default to elastic.")
def setlicense(license_type, elastic, password, user):
    license_type = license_type.lower()

    toolInstalled = is_tool('setlicense')

    if not toolInstalled:
        raise click.ClickException("""
            This command is a proxy for @pzl setlicense tool
            setlicense is not installed in your system.
            You can download it from https://github.com/elastic/pzl-es-tools/releases/
            Make it available in your $PATH and try again.
            """)

    subprocess.run(['setlicense', '-u', user, '-p',
                   password, '-e', elastic, license_type])



import click
from src.tmux import is_inside_tmux
from src.util import is_tool
import os
import sys


@click.command()
@click.argument("tag", required=True)
def checkout(tag):
    """
    Quickly creates a new context with the passed PR tag checked out,
    then bootstrap and starts kibbe setmeup

    """
    if not is_tool("gh"):
        click.echo("You must install github cli for this feature to work")
        exit(1)

    if not is_inside_tmux():
        click.echo("You must run this command inside a tmux session")
        exit(1)

    branch = tag.replace(":", "-")

    kibbe_command = sys.argv[0]
    os.system("git branch -D %s" % branch)
    os.system("%s ctx -i %s" % (kibbe_command, branch))
    os.system("gh pr checkout %s" % tag)
    os.system("%s setmeup --tmux" % kibbe_command)



from setuptools import setup, find_packages

setup(
    name="kibbe",
    packages=find_packages(),
    include_package_data=True,
    download_url="https://github.com/academo/kibbe/releases",
    url="https://github.com/academo/kibbe/",
    author="Esteban Beltran",
    author_email="kibbe@academo.me",
    install_requires=[
        "click",
        "termcolor",
        "libtmux",
        "requests",
        "psutil",
        "enlighten",
    ],
    entry_points={
        "console_scripts": [
            "kibbe = src.main:cli",
        ],
    },
)



import os
from pathlib import Path
import platform
from shutil import which
import subprocess
import random
import sys

import click
import libtmux
from termcolor import colored


def setup_tmux_client():
    if is_inside_tmux():
        # nothing to do
        return True

    tmux_installed = False

    try:
        tmux_bin = which("tmux")
        if tmux_bin and Path(tmux_bin).exists():
            tmux_installed = True
    except ValueError:
        pass

    if tmux_installed:
        click.echo(
            "You are not inside a tmux session. Create a new session first by"
            " running: \n"
        )

        tmux_command = ["tmux"]
        if platform.system() == "Darwin" and os.getenv("TERM_PROGRAM") == "iTerm.app":
            tmux_command.append("-CC")

        session_name = "kibbe_" + str(random.randint(0, 100))

        tmux_command = tmux_command + ["new-session", "-A", "-s", session_name]

        click.echo(colored(" ".join(tmux_command), "yellow"))

        if click.confirm("Do you want kibbe to run all for you?"):
            tmux_command.append("-d")
            subprocess.run(tmux_command)

            try:
                server = libtmux.Server()
                if not server.has_session(session_name):
                    raise Exception("no tmux session kibbe created")

                session = next(
                    session
                    for session in server.list_sessions()
                    if session.name == session_name
                )
                pane = session.attached_pane

                current_dir = os.getcwd()
                pane.send_keys("cd %s && nvm use" % current_dir)

                current_command = " ".join(sys.argv)
                pane.send_keys(current_command)

                # attach the current terminal to that session
                server.attach_session(session_name)
            except ValueError:
                raise click.ClickException(
                    colored(
                        "There was an error connecting to the tmux session. Check your"
                        " tmux installation and configuration",
                        "red",
                    )
                )

    else:
        click.echo("You don't have tmux installed. To install it:\n")

        if platform.system() == "Darwin":
            click.echo(colored("brew install tmux", "yellow"))
        elif platform.system() == "Linux":
            try:
                version = platform.version().lower()
                uname = platform.uname().release.lower()
                if "ubuntu" in version or "debian" in version:
                    click.echo(colored("sudo apt install tmux", "yellow"))
                elif "fedora" in version:
                    click.echo(colored("sudo yum install tmux", "yellow"))
                elif "manjaro" in uname or "arch" in uname:
                    click.echo(colored("sudo pacman -S tmux", "yellow"))
                else:
                    click.echo("Visit https://github.com/tmux/tmux/wiki/Installing")
            except Exception:
                click.echo("Visit https://github.com/tmux/tmux/wiki/Installing")

        # no tmux detected
        exit(1)


def is_inside_tmux():
    return os.getenv("TMUX") is not None


def get_current_panel():
    current_window = get_current_window()
    current_pane_id = os.getenv("TMUX_PANE")
    current_pane = current_window.get_by_id(current_pane_id)
    return current_pane


def get_current_window():
    tmux_panel = os.getenv("TMUX_PANE")
    tmux_session = os.getenv("TMUX")
    if not tmux_panel or not tmux_session:
        raise click.ClickException("Not inside a tmux session")

    tmux_session_id = "$" + tmux_session.split(",")[-1]
    server = libtmux.Server()

    session = server.get_by_id(tmux_session_id)
    if not session:
        raise click.ClickException("Could not interact with Tmux. Wrong tmux version")

    return session.attached_window


def clean_tmux_window(current_window):

    if len(current_window.panes) > 1:
        if not click.confirm(
            colored(
                "The current window has more than 1 pane. Do you want to close them all"
                " and start new ones?",
                "yellow",
            )
        ):
            raise click.ClickException(
                colored(
                    "Close all panes or create a new window to continue and run kibbe"
                    " again",
                    "red",
                )
            )
        current_pane_id = os.getenv("TMUX_PANE")
        current_pane = current_window.get_by_id(current_pane_id)
        panes = iter(current_window.panes)
        for pane in panes:
            if pane != current_pane:
                pane.cmd("kill-pane")



import re
import configparser
import os
from pathlib import Path
from src.util import get_valid_filename

import click
from termcolor import colored


class CustomConfigParser(configparser.ConfigParser):
    """Modified ConfigParser that allow ':' in keys and only '=' as separator."""

    OPTCRE = re.compile(
        r"(?P<option>[^=\s][^=]*)" r"\s*(?P<vi>[=])\s*" r"(?P<value>.*)$"
    )


def get_config_file(first=False):

    default_config_file = str(Path.home().joinpath(".kibbe"))

    config_files = []

    current_dir = os.getcwd()

    possible_paths = [
        default_config_file,
        str(Path.home().joinpath(".kibberc")),
        os.path.join(current_dir, ".kibbe"),
        os.path.join(current_dir, ".kibberc"),
    ]
    for path in possible_paths:
        possible_path = Path(path)
        if possible_path.exists():
            config_files.append(str(path))

    if len(config_files) == 0:
        config_files.append(default_config_file)

    if first:
        return config_files[0]

    return possible_paths


def make_config_files(config):
    for section in config.sections():
        if section.startswith("file-") and "content" in config[section]:
            try:

                current_dir = os.getcwd()
                filename = get_valid_filename(section[5:])
                filepath = os.path.join(current_dir, filename)
                content = config[section]["content"].lstrip()

                try:

                    # if the file exists, check they don't have different content
                    if Path(filepath).exists():
                        current_file = open(filepath, "r")
                        current_content = current_file.read()

                        # if the current content and future content are the same continue
                        if current_content == content:
                            continue
                        else:
                            if not click.confirm(
                                "File "
                                + colored(filename, "yellow")
                                + " already exists and has a different content."
                                " Overwrite?"
                            ):
                                continue
                except ValueError:
                    pass

                # write the config file
                click.echo("Writing config file " + colored(filename, "yellow"))
                file = open(filepath, "w")
                file.write(content)
                file.close()
            except ValueError:
                pass


def get_config():
    config = CustomConfigParser()
    try:
        config.optionxform = str  # type: ignore
    except ValueError:
        pass
    try:
        config.read(get_config_file())
    except ValueError:
        click.echo(colored("Error reading your configuration file.", "red"))
        pass
    return config


def get_kibbe_config(option: str = ""):
    config = get_config()

    if config.has_option("kibbe", option):
        return config.get("kibbe", option)

    return None


def persist_config(config_map):

    config = get_config_to_save(config_map)

    try:
        click.echo(colored("Config to save:\n ", "yellow"))
        print_config(config)
        if click.confirm(
            "Are you sure you want to save this configuration?\nAll existing"
            " configuration will be overwritten"
        ):
            config_file = open(str(get_config_file(first=True)), "w")
            config.write(config_file)
            click.echo(colored("Configuration written to kibbe config", "blue"))
        else:
            click.echo(colored("Cancelled", "yellow"))
            exit()
    except ValueError:
        exit()


def get_config_to_save(config_map):
    config = get_config()
    for config_key in config_map:
        if config_key not in config:
            config.add_section(config_key)

        data = config_map[config_key]
        if type(data) is dict:
            for key in data:
                wkey = key
                if wkey.startswith("--"):
                    wkey = wkey[2:]
                config.set(config_key, wkey, data[key])
        else:
            for value in data:
                value = value.split("=")
                if len(value) > 1:
                    config.set(config_key, value[0], value[1])
                else:
                    config.set(config_key, value[0], "")

    return config


def print_config(config):
    config_string = "--start--\n"
    sections = config.sections()
    for section in sections:
        config_string = config_string + "[%s]\n" % (section)
        items = config.items(section)
        for item, value in items:
            config_string = config_string + "%s = %s\n" % (item, value)
        config_string = config_string + "\n"
    config_string = config_string + "---end---"

    click.echo(config_string)
